import os
import shlex
from askGPT.tools import eprint, sanitizeName
def do_delete(shell, args):
    """delete: delete a subject."""
    args = shlex.split(args)
    if len(args) == 0:
        subject = shell.conversation_parameters["subject"]
    else:
        subject = sanitizeName(args[0])
    if subject in shell._config.get_list():
        os.remove(os.path.join(shell._config.conversations_path, subject + shell._config.fileExtention)) 
    else:
        eprint("Subject not found")
    shell._config.chat._chat_log = shell._config.chat._chat_log[:1]


def complete_delete(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    if not text:
        completions = shell._config.get_list()
    else:
        completions = [ f
                        for f in shell._config.get_list()
                        if f.startswith(text)
                        ]
    return completions


""""""
import os
from pathlib import Path
from .tools import load_json, eprint, strToValue
from askGPT import DATA_PATH
import toml
from askGPT.api.openai import ChatGPT

basicConfig = dict()
basicConfig["maxTokens"] = basicConfig.get("maxTokens",150)
basicConfig["model"] = basicConfig.get("model","gpt-3.5-turbo")
basicConfig["temperature"] = basicConfig.get("temperature",0.0)
basicConfig["topP"] = basicConfig.get("topP",1)
basicConfig["frequencyPenalty"] = basicConfig.get("frequencyPenalty",0.5)
basicConfig["presencePenalty"] = basicConfig.get("presencePenalty",0.5)
basicConfig["showDisclaimer"] = basicConfig.get("showDisclaimer",True)
basicConfig["maxRetries"] = basicConfig.get("maxRetries",3)
basicConfig["retryDelay"] = basicConfig.get("retryDelay",15.0)
basicConfig["retryMaxDelay"] = basicConfig.get("retryMaxDelay",60)
basicConfig["retryMultiplier"] = basicConfig.get("retryMultiplier",2)
basicConfig["verbose"] = basicConfig.get("verbose", False)
basicConfig["debug"] = basicConfig.get("debug", False)
basicConfig["updateScenarios"] = basicConfig.get("updateScenarios", True)

class Config(object):
    def __init__(self):
        self.rate_limit_per_minute = 20
        self.delay = 60.0 / self.rate_limit_per_minute
        self.disclaimer = "Disclaimer: The advice provided by askGPT is intended for informational and entertainment purposes only. It should not be used as a substitute for professional advice, and we cannot be held liable for any damages or losses arising from the use of the advice provided by askGPT."
        self.settingsPath = os.path.join(os.getenv("HOME"), ".askGPT")
        self.progConfig = dict()
        self.sessionConfig = dict()
        self.credentials = None
        self.api_base = None
        self.has = dict()
        self.has["license"] = False
        self.conversations_path=os.path.join(self.settingsPath, "conversations")
        Path(self.conversations_path).mkdir(parents=True, exist_ok=True)
        self.loadScenarios()
        self.fileExtention=".ai.txt"
        self.loadDefaults()
        self.loadProgConfig()
        self.update()
        self.chat = ChatGPT(self)
        self.chat.loadLicense()
        self.version="0.7.6"
        self.data_path = DATA_PATH

    def loadProgConfig(self):
        if os.path.isfile(os.path.join(self.settingsPath, "config.toml")):
            tomlConfig = toml.load(os.path.join(self.settingsPath,"config.toml"))
            self.progConfig.update(tomlConfig["default"])
        else:
            self.saveConfig()

    def updateParameter(self,key, val):
        val = strToValue(val)
        if key in self.sessionConfig: # order matters
            if self.sessionConfig[key] != val:
                print(f"{key}] = {val}")
                self.sessionConfig[key] = val
        elif key in self.progConfig:
            if self.progConfig[key] != val:
                print(f"{key}] = {val}")
                self.progConfig[key] = val
        


    def saveConfig(self):
        """Save the configuration file"""
        jsonConfig = {'name':'askGPT','default':self.progConfig}
        with open(os.path.join(self.settingsPath,"config.toml"), 'w') as f:
            toml.dump(jsonConfig,f)
        self.update()

    def reloadConfig(self):
        """Reload the configuration file"""
        self.update()

    def get_list(self):
        """
        list the previous conversations saved by askGPT."""
        conv_array = list()
        for line in os.listdir(self.conversations_path):
            if (not line.startswith("."))  and line.endswith(self.fileExtention) and (os.path.isfile(os.path.join(self.conversations_path,line))):
                conv_array.append(line.replace(self.fileExtention,""))
        return sorted(conv_array)

    def loadScenarios(self):
        """if there is not a file named scenarios.json, create it ad add the Neutral scenario"""
        if not os.path.isfile(os.path.join(self.settingsPath,"scenarios.json")):
            # copy the file from PATH
            with open(os.path.join(DATA_PATH,"scenarios.json"), "r") as f:
                data = f.read()
            with open(os.path.join(self.settingsPath,"scenarios.json"), "w") as f:
                f.write(data)
        self.scenarios = load_json(os.path.join(self.settingsPath,"scenarios.json"))


    def loadDefaults(self):
        self.progConfig["maxTokens"] = self.progConfig.get("maxTokens",150)
        self.progConfig["model"] = self.progConfig.get("model","text-davinci-003")
        self.progConfig["temperature"] = self.progConfig.get("temperature",0.0)
        self.progConfig["topP"] = self.progConfig.get("topP",1)
        self.progConfig["frequencyPenalty"] = self.progConfig.get("frequencyPenalty",0.0)
        self.progConfig["presencePenalty"] = self.progConfig.get("presencePenalty",0.0)
        self.progConfig["showDisclaimer"] = self.progConfig.get("showDisclaimer",True)
        self.progConfig["maxRetries"] = self.progConfig.get("maxRetries",3)
        self.progConfig["retryDelay"] = self.progConfig.get("retryDelay",15.0)
        self.progConfig["retryMaxDelay"] = self.progConfig.get("retryMaxDelay",60)
        self.progConfig["retryMultiplier"] = self.progConfig.get("retryMultiplier",2)
        self.progConfig["verbose"] = self.progConfig.get("verbose", False)
        self.progConfig["debug"] = self.progConfig.get("debug", False)
        self.progConfig["updateScenarios"] = self.progConfig.get("updateScenarios", True)

    def printConfig(self):
        """Print the configuration file"""
        print(toml.dumps(self.progConfig))

    def update(self):
        """
Load the configuration file from ~/.askGPT/config.toml"""
        self.loadProgConfig()
            # self.progConfig.update(tomlConfig["askGPT"])



import shlex
from askGPT.tools   import eprint
import os

def do_recap(shell, args):
    """prints the text from the conversation file using the subject.ai.txt """
    args = shlex.split(args)
    if len(args) == 0:
        subject = shell.conversation_parameters["subject"]
    else:
        subject = args[0]
        if subject not in shell._config.get_list():
            eprint(f"Subject {subject} not found")
            return
    filename = os.path.join(shell._config.conversations_path, f"{subject}{shell._config.fileExtention}")
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            print(f.read())
    else:
        eprint(f"File {filename} not found")


def complete_recap(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    if not text:
        completions = shell._config.get_list()
    else:
        completions = [ f
                        for f in shell._config.get_list()
                        if f.startswith(text)
                        ]
    return completions


import openai
import os

from askGPT.tools import eprint, sanitizeName
import time
import backoff
import click

"""This is a class that inherit from openai class that will allow us to query chatgpt. By using a class we can share the object between modules passing it as an argument."""
class ChatGPT(object):
    def __init__(self, config) -> None:
        super().__init__()
        self._model = "gpt-3.5-turbo"
        self._temperature = 0.9
        self._max_tokens = 150
        self._top_p = 1
        self._frequency_penalty = 0
        self._presence_penalty = 0
        # self._stop = ["\n"]
        self._config = config
        self._chat_log = []


    def listModels(self):
        models = list()
        for model in sorted(list(map(lambda n: n.id,openai.Model.list().data))):
            models.append(model)
        return models
       
    def editDialog(self,subject):
        """
        Edit a conversation"""
        subject = sanitizeName(subject)
        lines = list()
        if os.path.isfile(os.path.join(self._config.conversations_path, subject + self._config.fileExtention)):
            with open(os.path.join(self._config.conversations_path, subject + self._config.fileExtention), "r") as f:
                lines = f.readlines()
        
        lines = click.edit("".join(lines))
        if lines is not None:
            with open(os.path.join(self._config.conversations_path, subject + self._config.fileExtention), "w") as f:
                f.write(lines)
                self.createPrompt(subject, None, None)
        else:
            eprint("No changes made")



    def bootStrapChat(self,scenario):
        """Read the scenario and return the initial chat"""
        chat = list()
        
        if scenario in self._config.scenarios:
            self.greetings = {"role" : "system", "content":  self._config.scenarios[scenario]["greetings"] }
            conversationChat = [self.greetings]
            return  conversationChat + self._config.scenarios[scenario]["conversation"]
        else:
            eprint("Scenario not found")
            return []

    def completions_with_backoff(self, delay_in_seconds: float = 1,**kwargs):
        """Delay a completion by a specified amount of time."""
        # Sleep for the delay
        time.sleep(delay_in_seconds)
        if self._config.progConfig.get("api_base",None) is not None:
            openai.api_base = self._config.progConfig["api_base"]
        return openai.ChatCompletion.create(**kwargs)

    def createPrompt(self, subject: str, scenario: str, enquiry: dict):
        subject = sanitizeName(subject)
        chat = list()
        if subject:
            with open(os.path.join(self._config.conversations_path, sanitizeName(subject) + self._config.fileExtention), "a") as f:
                pass
            with open(os.path.join(self._config.conversations_path, sanitizeName(subject) + self._config.fileExtention), "r") as f:
                chatRaw = f.readlines()
                bootstrappedChat = list()
                if scenario:
                    bootstrappedChat = self.bootStrapChat(scenario)
                for line in chatRaw:
                    if line.startswith("user:"):
                        bootstrappedChat.append({"role": "user", "content": line.replace("user: ","")})
                    elif line.startswith("assistant:"):
                        bootstrappedChat.append({"role": "assistant", "content": line.replace("assistant: ","")})
                    elif line.startswith("system:"):
                        bootstrappedChat.append({"role": "system", "content": line.replace("system: ","")})
                    else:
                        bootstrappedChat[-1]["content"] += line
            
                """we need to add the enquiry to the chat"""
                if enquiry:
                    bootstrappedChat.append(enquiry)
                
                """We return a list that concats bootstrappedChat andchat"""
                return bootstrappedChat
        else:
            eprint("Please set a subject")
            return []


    def query(self, subject: str, scenario: str, enquiry: str, max_tokens: int = 150, temperature: float = 0.9, top_p: float = 1, frequency_penalty: float = 0, presence_penalty: float = 0, stop: list = ["\n", " user:", " assistant:"]):
        """Query the model with the given prompt."""
        # Load the license
        if not self.loadLicense():
            return
        # Create the prompt
        
        # We will prepend a system prompt with the information gathered from the me.txt file if file exists in the .askGPT config directory
        # open the ~/.askGPT/me.txt
        prompt = ""
        if self._config.progConfig.get("useMemoryFile",False):
            try:
                meFilePath = os.path.join(self._config.settingsPath,self._config.progConfig.get("memoryFile"))
                meText = ""
                if os.path.exists(meFilePath):
                    with open(meFilePath,'r', encoding='utf-8') as file:
                        meText = file.read().strip()
                        if meText != "":
                            meText = "\n\nUser info:\n"+meText+"\n\n"
                            prompt = {"role": "system", "content": meText}
            except Exception as ex:
                eprint ("Error reading me.txt : "+str(ex))
        # if prompt has a value then we init chat with it and then append list(self._chat_log)
        if prompt != "":
            chat = [prompt]+list(self._chat_log)
        else:
            chat  = list(self._chat_log)
        chat.append({"role":"user", "content": enquiry})
        # print("sending chat:")
        # print(chat)
        # return 
        ai = self.submitDialogWithBackOff(chat)
        if ai:
            # Add the response to the chat log
            self._chat_log.append({"role": "user", "content": enquiry})
            self._chat_log.append({"role": "assistant", "content": ai})
            # Return the response
            return ai

    def saveLicense(self, api_key):
        if not os.path.isdir(self._config.settingsPath):
            os.mkdir(self._config.settingsPath)
        with open(os.path.join(self._config.settingsPath, "credentials"), "w") as f:
            f.write(api_key)
        return True

    def load(self,chat):
        self.greetings = chat[0]
        self._chat_log = chat[1:]

    def loadLicense(self):
        # Load your API key from an environment variable or secret management service
        if os.getenv("OPENAI_API_KEY"):

            openai.api_key = os.getenv("OPENAI_API_KEY")
            self._config.credentials = os.getenv("OPENAI_API_KEY")
            self._config.has["license"] = True
            return True
        else:
            if os.path.isfile(os.path.join(self._config.settingsPath, "credentials")):
                with open(os.path.join(self._config.settingsPath, "credentials"), "r") as f:
                    credentials = f.read()
                    openai.api_key = credentials.strip()
                    self._config.credentials = credentials
                    self._config.has["license"] = True
                    return True
            else:
                eprint("Please set OPENAI_API_KEY and OPENAI_ORGANIZATION environment variables.")
                eprint("Or create a file at ~/.askGPT/credentials with the following format:")
                eprint("OPENAI_API_KEY:OPENAI_ORGANIZATION")
            
        return False
    
    def dream(self, prompt):
        response = openai.Image.create(
        prompt=prompt,
        n=1,
        size="1024x1024"
        )
        image_url = response['data'][0]['url']
        return image_url

    def submitDialog(self, subject, scenario):
        """Send the dialog to openai and save the response"""
        subject = sanitizeName(subject)
        chat = ""
        if subject:
            with open(os.path.join(self._config.conversations_path, sanitizeName(subject) + self._config.fileExtention), "a") as f:
                pass
            #chat = self.createPrompt(subject, scenario, None)
            chat = list(self._chat_log)
            if chat == None:
                print("Empty conversation")
                return
            ai = self.submitDialogWithBackOff(chat)
            if ai:
                return ai

    def submitDialogWithBackOff(self, chat):
        tries = self._config.progConfig.get("maxRetries",1)
        success = False
        reason = "Error: Could not send the dialog"
        sleepBetweenRetries = self._config.progConfig["retryDelay"]
        ai  = ModuleNotFoundError
        while tries > 0:
            try:
                if self._config.progConfig["debug"]:
                    eprint(chat)
                conversation = list(chat)
                conversation.insert(0, self.greetings)
                response = self.completions_with_backoff(
                    delay_in_seconds=self._config.delay,
                    model=self._config.progConfig["model"],
                    messages=conversation,
                    temperature=self._config.progConfig["temperature"],
                    max_tokens=self._config.progConfig["maxTokens"],
                    top_p=self._config.progConfig["topP"],
                    frequency_penalty=self._config.progConfig["frequencyPenalty"],
                    presence_penalty=self._config.progConfig["presencePenalty"],
                
                )
                # print(response)
                # return
                ai = response.choices[0]['message'].content
                if ai.startswith("\n\n"):
                    ai = ai[2:]
                if self._config.progConfig["debug"]:
                    eprint(ai)
                success = True
                break
            except KeyboardInterrupt:
                reason = "Operation aborted."
                break
            except Exception as e:
                tries -= 1
                if str(e) == "openai.error.RateLimitError":
                    eprint("Error: Too many requests. We will try again")
                if str(e).startswith("This model's maximum context length is"):
                    eprint("Error: Too many tokens. We will try again with less history")
                    eprint(f"Current number of interactions: {len(chat)}")
                    chat = chat[int((len(chat)/round(tries + 0.51)) + 0.5):]
                    self._chat_log = chat
                    eprint(f"New number of interactions: {len(chat)}")
                    time.sleep(5)
                    continue
                
                eprint("Error: " + str(e))

                eprint(f"Retrying again in {sleepBetweenRetries} seconds...")
                time.sleep(sleepBetweenRetries)
                sleepBetweenRetries *= self._config.progConfig["retryMultiplier"] 
                if sleepBetweenRetries > self._config.progConfig["retryMaxDelay"]:
                    sleepBetweenRetries = self._config.progConfig["retryMaxDelay"]
        if success == False:
            eprint(reason)
            return
        return ai

    def get_chat_log(self):
        """Get the chat log."""
        return self._chat_log

    def set_model(self, model):
        """Set the model."""
        self._model = model

    def set_temperature(self, temperature):
        """Set the temperature."""
        self._temperature = temperature

    def set_max_tokens(self, max_tokens):
        """Set the max tokens."""
        self._max_tokens = max_tokens
    
    def set_top_p(self, top_p):
        """Set the top p."""
        self._top_p = top_p

    def set_frequency_penalty(self, frequency_penalty):
        """Set the frequency penalty."""
        self._frequency_penalty = frequency_penalty

    def set_presence_penalty(self, presence_penalty):
        """Set the presence penalty."""
        self._presence_penalty = presence_penalty

    def set_stop(self, stop):
        """Set the stop."""
        self._stop = stop

    def get_model(self):
        """Get the model."""
        return self._model

    def get_temperature(self):
        """Get the temperature."""
        return self._temperature

    def get_max_tokens(self):
        """Get the max tokens."""
        return self._max_tokens

    def get_top_p(self):
        """Get the top p."""
        return self._top_p

    def get_frequency_penalty(self):
        """Get the frequency penalty."""
        return self._frequency_penalty

    def get_presence_penalty(self):
        """Get the presence penalty."""
        return self._presence_penalty

    def get_stop(self):
        """Get the stop."""
        return self._stop









def do_me(shell, args):
    import os
    from askGPT.tools import eprint    
    filename = os.path.join(shell._config.settingsPath, shell._config.progConfig["memoryFile"])
    # Check for existence of file and create it if not there
    if not os.path.exists(filename):
        open(filename, 'w').close()
    else:
        pass
    if len(args) == 0:
        eprint("/me <show|add|del>")
    elif args == "show":
        try:
            with open(filename, 'r') as filehandle:
                lines = filehandle.readlines()
                print("\n".join([line.strip('\n') for line in lines]))
        except Exception as e:
            shell.print_exception(e)
    elif args.startswith("add"):
            text = args[4:]
            try:
                with open(filename, 'a+') as filehandle:
                    filehandle.write(f'\n{text}')
            except Exception as e:
                shell.print_exception(e)
    elif args.startswith("del"):
        text = args[4:]
        eprint(f"deleting line starting with:\n{text} = ")
        try:
            with open(filename, 'r') as filehandle:
                lines = filehandle.readlines()
                # we will filter out lines starting with arg[1] = ...
                filtered_lines = [line for line in lines if line.startswith(f"{text} = ") is False]
            with open(filename, 'w') as filehandle:
                filehandle.writelines(filtered_lines)
        except Exception as e:
            shell.print_exception(e)
    else:
        eprint(f"Invalid arguments: {args}")


import os
import shlex
from askGPT.tools import eprint

def do_exec(shell,args):
        """Execute a the rest of the line in a bash shell and print the output"""
        args = shlex.split(args)
        if len(args) == 0:
            eprint("No command provided")
            return
        command = " ".join(args)
        os.system(command)


import shlex
from askGPT.tools   import eprint, sanitizeName
import toml

def do_show(shell, arg):
    """
    show: show the config|scenarios|models|subjects or the conversation inside a subject.
    <config|scenarios|models|subjects|subject <subject>>"""
    args = shlex.split(arg)
    if len(args) == 0:
        eprint("Show config|scenarios|models|subjects or the conversation inside a subject.")
        return
    elif len(args) == 1:
        if args[0] == "config":
            print("Current configuration:")
            print(toml.dumps(shell._config.progConfig))
            return
        elif args[0] == "scenarios":
            print("Current scenarios:")
            for scenario in sorted(shell._config.scenarios.keys()):
                print(scenario)
            return
        elif args[0] == "subjects":
            print("Current subjects:")
            for subject in shell._config.get_list():
                print(subject)
            return
        elif args[0] == "models":
            if shell._config.has.get("license", False):
                print("Current models:")
                for val  in shell._config.chat.listModels():
                    print(val)
            else: 
                shell._config.chat.loadLicense()
            return
        else:
            if shell.conversation_parameters.get("defaultCommand", "") == "query":
                shell.commands["query"](arg)
            else:
                eprint("Unrecognized parameter.")
            return
    elif len(args) == 2:
        if args[0] == "scenario":
            scenario = sanitizeName(args[1])
            scenario = shell._config.scenarios.get(scenario)
            if scenario is not None:
                print("Greetings:")
                print(scenario.get("greetings", ""))
                for prompt in scenario["conversation"]:
                    print("{} {}".format(scenario.get(prompt["role"], shell._config.progConfig.get(prompt["role"])), prompt["content"]))
    else:
            if shell.conversation_parameters.get("defaultCommand", "") == "query":
                shell.commands["query"](arg)
                return
            else:
                eprint("Unrecognized parameter.")
                return
        
def complete_show(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    args = shlex.split(line)
    completions = None
    # print(f"{args}\n")
    if len(args) < 2:
        if not text:
            completions = list(["config", "scenarios", "subject", "subjects"] )
        else:
            completions = [ f
                            for f in list(["config", "scenarios", "subject", "subjects"] )
                            if f.startswith(args[-1])
                            ]
    elif len(args) == 2:
        if args[1] == "scenario":
            completions = list(shell._config.scenarios.keys())
        else:
            completions = [ f
                            for f in list(["config", "scenarios", "scenario", "subjects"] )
                            if f.startswith(args[-1])
                            ]
    elif len(args) == 3:
        completions = [ f
                        for f in list(shell._config.scenarios.keys())
                        if f.startswith(args[2])
                        ]
    return completions


import shlex
from askGPT.tools   import eprint, sanitizeName
import os

def do_clone(shell,args):
    """if len(arg) == 1 then copy the current conversation to a new file using arg[1]"""
    args = shlex.split(args)
    if len(args) == 0:
        eprint("You must provide a name for the new conversation")
    else:
        if len(args) == 1:
            new_subject = sanitizeName(args[0])
        else:
            eprint("You can only provide one argument")
            return
        if new_subject in shell._config.get_list():
            eprint(f"Subject {new_subject} already exists")
            return
        current_subject = shell.conversation_parameters["subject"]
        filename = os.path.join(shell._config.conversations_path, f"{current_subject}{shell._config.fileExtention}")
        with open(filename, "r") as r:
            text = r.read()
        filename = os.path.join(shell._config.conversations_path, f"{new_subject}{shell._config.fileExtention}")
        with open(filename, "w") as w:
            w.write(text)
        shell.conversation_parameters["subject"] = new_subject
        print(f"Conversation {new_subject} created")


import shlex
"""capiche
command that will submit the chat plus the Human entry: If you understand say so.
"""
def do_capiche(shell, args):
    """capiche will submit the chat plus the Human entry: If you understand say so."""
    args = shlex.split(args)
    capiche = "If you understand the task, don't do it, just respond 'Capiche!' and nothing else."
    if len(args) == 0:
        shell.do_query(capiche)
    else:
        args.append(".")
        args.append(f"{capiche}")
        shell.do_query(" ".join(args))


import shlex
from askGPT.tools import sanitizeName, eprint, strToValue

def do_set(shell,args):
    """Set conversation_parameters checking that the keys exist and that the values are valid"""
    args = shlex.split(args)
    if len(args) == 0:
        for key in shell.conversation_parameters:
            print(f"{key} = {shell.conversation_parameters[key]}")
        return
    if len(args) == 2:
        key = args[0]
        val = strToValue(args[1])
        if key in shell.conversation_parameters:
            if key == "scenario":
                if val in shell._config.scenarios:
                    shell.conversation_parameters[key] = val
                    shell.prompt = f"{val}> "
                    shell._config.chat.greetings = {"role": "system", "content": shell._config.scenarios[val]["greetings"]}
                else:
                    eprint("Scenario not found")
            elif key == "model":
                if val  in shell._config.chat.listModels():
                    shell.conversation_parameters[key] = val
                else:
                    eprint("Model not found")
            else:
                shell.conversation_parameters[key] = val
        elif key in shell._config.progConfig:
            if key == "fileExtention":
                if val[0] != ".":
                    val = "." + val
            shell._config.progConfig[key] = strToValue(val)
            shell._config.saveConfig()
        else:
            eprint("Key not found")
    else:
        eprint("Wrong number of arguments")

def complete_set(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    # print("len: {}, spaces: {}\n".format(len(line), len(line.split(" "))))
    if  line.rstrip()  == "set":
        completions = list(shell.conversation_parameters.keys())
        return completions
    elif len(line.split(" ")) == 2:
        completions = [ f
                        for f in shell.conversation_parameters.keys()
                        if f.startswith(text)
                        ]
        return completions
    elif len(line.split(" ")) >  2:
        completions = line.rstrip().split(" ")[1]
        if completions == "scenario":
            completions = [
                f
                for f in list(shell._config.scenarios.keys())
                if f.startswith(text) 
            ]
            return completions
        elif completions == "subject":
            completions = [
                f
                for f in list(shell._config.get_list())
                if f.startswith(text) 
            ]
            return completions
        elif completions == "model":
            completions = [
                f
                for f in list(shell._config.chat.listModels())
                if f.startswith(text) 
            ]
            return completions
        elif completions == "defaulCommand":
            completion =  [
                f
                for f in list(shell.commands.keys())
                if f.startswith(text)
            ]
            return completions



from askGPT.tools import eprint

def do_continue(shell, args):
    """This is in case we get an exception that we didn't protect ourselves from"""
    if args:
        eprint(args)
    else:
        eprint("An unknown error occured.")


import cmd
import shlex
import pkgutil
import shutil 
from .tools   import eprint, sanitizeName
import toml
import os
import sys
import rich
import importlib
from pathlib import Path
from rich import print
from rich.text import Text
from rich.style import Style
from rich.prompt import Prompt, Confirm
from rich.panel import Panel
from rich.markdown import Markdown
from .tools import strToValue, addMetadata
import requests
import datetime
from rich.console import Console
import click
from filecmp import cmp
import subprocess

danger_style = Style(color="red", blink=False, bold=True)
attention_style = Style(color="yellow", blink=False, bold=True)
ok_style = Style(color="green", blink=False, bold=False)



"""Here we will define the class Shell which is a child of cmd.cmd which will allow us to run all the commands interactively such as query, config, edit."""
class Shell(cmd.Cmd):
    def __init__(self, config) -> None:
        super().__init__()
        self.console = Console()
        self.intro = "Welcome to askGPT. Type help or ? to list commands."
        self.doc_header = "Commands (type help <topic>):"
        self.misc_header = "Miscellaneous help topics:"
        self.undoc_header = "Undocumented commands:"
        self.ruler = "-"
        self._config = config
        self.commands = dict()
        self.lastResponse = None
        self._register_commands()
        self._register_completion_methods()
        self.conversation_parameters = {
            "subject": "test",
            "scenario": "ChatGPT",
            "model": "gpt-3.5-turbo",
            "defaultCommand": "query",
            "execute": False

        }
        if os.path.exists(os.path.join(self._config.settingsPath, "last.toml")):
            self.conversation_parameters.update(toml.load(os.path.join(self._config.settingsPath, "last.toml")))
        self.prompt = f"{self.conversation_parameters['scenario']}> "
        # when we load we initializr the chat list
        self.chatList = self._config.chat.createPrompt(self.conversation_parameters['subject'], self.conversation_parameters['scenario'], None)
        self._config.chat.load(self.chatList)
        self.commands["update"]("")

    
    def _register_commands(self):
        """Register all the commands."""
        command_package = 'askGPT.commands'  # This is your package where commands are located.
        
        # Dynamically list the modules in the specified package.
        command_modules = pkgutil.iter_modules(importlib.import_module(command_package).__path__)

        for module_info in command_modules:
            if module_info.name.endswith('_command'):
                command_name = "unknown"
                try:
                    # Import the module
                    module = importlib.import_module(f".{module_info.name}", command_package)

                    # Extract the command's name from the module name (assuming the format is '<command>_command')
                    command_name = module_info.name.rsplit('_', 1)[0]

                    # Find and register 'do_' functions (i.e., the command actions)
                    do_function = getattr(module, f"do_{command_name}", None)
                    if callable(do_function):
                        # Bind the function as a method of this class instance
                        bound_method = do_function.__get__(self)
                        setattr(self, f"do_{command_name}", bound_method)
                        self.commands[command_name] = bound_method  # Optional, if you want to keep track of commands
                        
                    # Here, you can also extend it to find and register 'help_' and 'complete_' functions, if available.

                except ImportError as e:
                    # Handle import errors here. You can log the error message if you have a logging system.
                    print(f"Failed to import command '{command_name}': {str(e)}")


    def _register_completion_methods(self):
        # Assuming 'askGPT.shell' is the package where Shell class is defined,
        # and 'commands' is a subpackage containing your commands.
        command_package = 'askGPT.commands'

        # Get the path of the 'commands' subpackage
        package_path = importlib.import_module(command_package).__path__

        # Discover and load all command modules in the 'commands' subpackage
        for importer, modname, ispkg in pkgutil.iter_modules(package_path):
            try:
                # Import the module
                module = importlib.import_module(f".{modname}", command_package)
            except ImportError as e:
                print(f"Error importing module {modname}: {e}")
                continue  # Skip this module

            # Attach complete_* functions from the module to the Shell instance
            for attribute_name in dir(module):
                if attribute_name.startswith("complete_"):
                    # Get the function from the module
                    complete_function = getattr(module, attribute_name)

                    # Bind the function as a method of this class instance
                    bound_method = complete_function.__get__(self)
                    setattr(self, attribute_name, bound_method)
        
    

    def precmd(self, line):
        """This method is called after the line has been input but before
        it has been interpreted. If you want to modify the input line
        before execution (for example, variable substitution) do it here.
        """
        if line.startswith('/'):
            return line[1:]  # Remove the slash and treat the rest as the command
        elif line.startswith("!"):
            return line
        elif line.startswith('?'):
            return line
        else:
            return "query " + line
        

    def postcmd(self, stop, line):
        """postcmd: print the result."""
        #print("Result:", stop)
        return stop


    """Write function similar to do_exec but that it is able to save the output from the command being executed."""
    def execStdout(self, command: list):
        result = subprocess.run(args=command, stdout=subprocess.PIPE)
        return result.stdout.decode('utf-8')
    
    
    def saveSession(self):
        """EOF: exit the shell."""
        """ save the current parameters in a toml file to be loaded in the next session"""
        with open(os.path.join(os.path.join(self._config.settingsPath, "last.toml")), "w") as f:
            toml.dump(self.conversation_parameters, f)


    def emptyline(self):
        """emptyline: do nothing."""
        pass

    def default(self, line):
        if line == "!!":
            # take the last response from AI and esecute it. This is meanly useful when using the sharedTermninal scenario. We want to capture the output and feed it into the conversation as the user prompt.
            if self.lastResponse:
                self.commands["exec"](self.lastResponse)

            pass
        elif line.startswith("!"):
            self.commands["exec"](line[1:])
        elif self.conversation_parameters.get("defaultCommand","query") == "query":

            self.commands["exec"](line)
        else:
            """default: print the error message."""
            print("Unrecognized command:", line)

    def postloop(self):
        """postloop: print the exit message."""
        print("Exiting askGPT.")

    



    



import shlex
from askGPT.tools   import eprint

"""Show the greeting and the conversation prompt used to precondition the scenario"""
def do_intro(shell, args):
    """if args is one of the scenarios, print the greeting of that scenario"""
    args = shlex.split(args)
    scenario = shell.conversation_parameters["scenario"]
    if len(args) != 0:
        scenario = args[0]
        """print current scenario from conversation_parameters"""
    if scenario in shell._config.scenarios:    
        print(f"system: {shell._config.scenarios[scenario]['greetings']}")
        for p in shell._config.scenarios[scenario]['conversation']:
            print(f"{p['role']}: {p['content']}")
            
    else:
        eprint(f"Scenario {scenario} not found")


def do_help(shell, arg):
    """help: show the help message."""
    if arg:
        try:
            print(getattr(shell, 'do_' + arg).__doc__)
        except AttributeError:
            print(shell.nohelp % (arg,))
    else:
        names = shell.commands.keys()
        cmds_doc = []
        cmds_undoc = []
        cmds_misc = []
        for name in names:
            if name[:5] == 'help_':
                cmds_misc.append(name[5:])
            elif shell.commands[name].__doc__:
                cmds_doc.append(name)
            else:
                cmds_undoc.append(name)
        shell.print_topics(shell.doc_header, cmds_doc, 15, 80)
        shell.print_topics(shell.misc_header, cmds_misc, 15, 80)
        shell.print_topics(shell.undoc_header, cmds_undoc, 15, 80)


import shlex
import os
from filecmp import cmp
from ..tools   import eprint, sanitizeName
from rich.prompt import Prompt, Confirm
import shutil

def do_update(shell,args):
    """replace the current scenarios files"""
    args = shlex.split(args)
    if shell._config.progConfig["updateScenarios"] == False:
        eprint("Update is disabled in the config file")
        return
    if not cmp(os.path.join(shell._config.data_path, "scenarios.json"), os.path.join(shell._config.settingsPath, "scenarios.json")):
        if not Confirm.ask("New scenarios available.Would you like to replace the current ones?"):
            eprint("Scenarios files matched. No need to overwrite.")
            return 
        if Confirm.ask("Would you like to make a backup of the current one?"):
            shutil.copyfile(os.path.join(shell._config.settingsPath, "scenarios.json"), os.path.join(shell._config.settingsPath, "scenarios.json.bak"))
        shutil.copyfile(os.path.join(shell._config.data_path, "scenarios.json"), os.path.join(shell._config.settingsPath, "scenarios.json"))
        eprint("Scenarios updated, you need to restart to load the new scenario file")
        return 


import shlex
from askGPT.tools import eprint, strToValue
def do_config(shell, args):
    """needed two parameters, the first one is the key and the second one is the value"""
    args = shlex.split(args)
    clean = True
    if len(args) == 0:
        shell._config.printConfig()
        return
    elif len(args) == 1:
        if args[0] == "save":
            shell._config.saveConfig()
            return
        eprint("No value provided.")
        return
    elif len(args) == 2:
        value = strToValue(args[1])
        if args[0] in shell._config.progConfig:
            shell._config.progConfig[args[0]] = value
        elif args[0] in shell._config.conversation_parameters:
            shell._config.updateParameter(args[0], value)
    else:
        eprint("Unrecognized parameter.")
    if not clean:
        pass

def complete_config(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    if not text:
        completions = list(shell._config.progConfig.keys())
    else:
        completions = [ f
                        for f in shell._config.progConfig.keys()
                        if f.startswith(text)
                        ]
    return completions



import os
from rich.markdown import Markdown
import importlib.resources as pkg_resources
from askGPT.tools import eprint

def do_man(shell, line):
    """ when passed the name of a command it will pint an extensive explanation of the command by loading a markup file from data/"""
    if line:
        command = shell.parseline(line)
        if command[0] in shell.commands:
            command = command[0]
        else:
            command = None
    else:
        command = None
    if command:
        # To get the path to a directory 'data' within your 'askGPT' package
        try:
            # This gives you a path-like object you can use
            data_path = pkg_resources.files('askGPT').joinpath('data')
            data_dir = str(data_path)
            man_path = os.path.join( data_dir, "docs", "man_" + command + ".md")
            eprint(f"man_path = {man_path}")
            with open(man_path, "r") as f:
                shell.console.print( Markdown(f.read()))
        except FileNotFoundError:
            print("No manual entry for", command)   

def complete_man(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    if not text:
        completions = list(shell.commands.keys())
    else:
        completions = [ f
                        for f in shell.commands.keys()
                        if f.startswith(text)
                        ]
    return completions


import shlex
from askGPT.tools import sanitizeName, eprint

def do_edit(shell, args):
        args = shlex.split(args)
        """edit: edit a subject."""
        if len(args) == 0:
            subject = shell.conversation_parameters["subject"]
        else:
            subject = sanitizeName(args[0])
        if subject in shell._config.get_list():
            shell._config.chat.editDialog(subject)
        else:
            eprint("Subject not found")

def complete_edit(shell,text, line, begidx, endidx):
    """complete_query: complete the query command."""
    if not text:
        completions = shell._config.get_list()
    else:
        completions = [ f
                        for f in shell._config.get_list()
                        if f.startswith(text)
                        ]
    return completions


import shlex
import datetime
import os
from askGPT.tools import eprint, addMetadata
import requests
"""dream will help you generate images based on your prompt"""
def do_dream(shell, args):
    args = shlex.split(args)
    if len(args) == 0:
        eprint("No prompt provided.")
        return
    prompt = " ".join(args)
    """ show rich progress"""
    with shell.console.status("waiting for response ...", spinner="dots"):
        url = shell._config.chat.dream(prompt)
        if url:
            """ download the image into the conversation directory using the <subject>_<date>  add the prompt to the metadata of the image after saving. """
            subject = shell.conversation_parameters["subject"]
            date = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            filename = f"{subject}_{date}.png"
            filepath = os.path.join(shell._config.conversations_path, filename)
            with open(filepath, "wb") as f:
                f.write(requests.get(url).content)
                addMetadata(filepath,  f"{prompt}")  
                print(f"Image saved at {filepath}")
        else:
            eprint("Error while generating the image.")
        



import shlex
from askGPT.tools import eprint

def do_greetings(shell, args):
    """if args is one of the scenarios, print the greeting of that scenario"""
    args = shlex.split(args)
    if len(args) == 0:
        """print current scenario from conversation_parameters"""
        scenario = shell.conversation_parameters["scenario"]
        print(f"system: {shell._config.scenarios[scenario]['greetings']}")
        return
    scenario = args[0]
    if scenario in shell._config.scenarios:
        print(f"system: {shell._config.scenarios[scenario]['greetings']}")
    else:
        eprint(f"Scenario {scenario} not found")



import os
import shlex
from askGPT.tools import eprint
from filecmp import cmp
from ..tools   import eprint, sanitizeName
from rich.prompt import Prompt, Confirm

def do_credentials(shell, args):
    """credentials: show the credentials."""
    if shell._config.credentials:
        print(shell._config.credentials)
        if not Confirm.ask("Would you like to replace them?"):
            return
    """ask if to replace"""
    api_key = ""
    while(api_key == ""):
        api_key = Prompt.ask("Enter your openai API key",default="" if shell._config.credentials is None else shell._config.credentials.split(":")[0])
    shell._config.credentials = f"{api_key}"

    """if yes, ask for the new credentials"""
    shell._config.chat.saveLicense(api_key)
    shell._config.chat.loadLicense()
    """if no, do nothing"""



import sys
import json

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def addMetadata(path, text):
    """Add Exif metadata to png file (path)"""
    from PIL import Image
    from PIL.ExifTags import TAGS
    img = Image.open(path)
    exifdata = img.getexif()
    if exifdata is None:
        exifdata = {}
    exifdata[270] = text
    img.save(path, exif=exifdata)


def sanitizeName(name):
    """
    Sanitize the name of the conversation to be saved."""
    return name.replace(" ", "_").replace("/", "_")

def load_json(file):
    """
    Load json from file"""
    with open(file, "r") as f:
        try:
            return json.load(f)
        except:
            return dict()

def strToValue(val):
    if val == "true":
        val = True
    elif val == "false":
        val = False
    elif val.isnumeric():
        val = int(val)
    elif val.replace(".","",1).isnumeric():
        val = float(val)
    return val


import os
from filecmp import cmp
from askGPT.tools   import eprint, sanitizeName
from rich.prompt import Prompt, Confirm
from rich.text import Text
import click
import subprocess

def do_query(shell, enquiry):
    """Query the model with the given prompt."""
    """query: query the model with the given prompt.
        <prompt> """
    if not shell._config.has.get("license", False):
        shell._config.chat.loadLicense()
        return
    if shell._config.has.get("license", False):
        response = None
        with shell.console.status("waiting for response ...", spinner="dots"):
            response = shell._config.chat.query(shell.conversation_parameters["subject"], shell.conversation_parameters["scenario"], enquiry)
        if response:
            text = Text(response)
            text.stylize("bold magenta")
            shell.console.print(f"{text}\n")
            shell.lastResponse = text
            """save to file"""
            with open(os.path.join(shell._config.conversations_path, shell.conversation_parameters["subject"] + shell._config.fileExtention), "a") as f:
                    if shell.conversation_parameters.get("execute", False):
                            editPrompt = click.prompt(f"edit command? [y/n]", type=click.Choice(["y", "n"]), default="n")
                            if editPrompt == "y":
                                edited = click.edit(response)
                                if edited:
                                    response = edited
                            doExec = click.prompt(f"{response}\nExecute command? [y/n]", type=click.Choice(["y", "n"]), default="y")
                            """execute the command in the terminal and edit the response before saving it."""
                            if doExec == "y":
                                result  = subprocess.run(response, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
                                result = result.stdout.decode("utf-8")
                                print(result)
                                saveOutput = click.prompt(f"save output? [Y/e/n]", type=click.Choice(["y", "e", "n"]), default="y")
                                if saveOutput == "e":
                                    edited = click.edit(result)
                                    if edited:
                                        result = edited
                                if saveOutput != "n":
                                    f.write(f"user: {str(enquiry)}")
                                    f.write("\n")
                                    f.write(f"assistant: {response}")
                                    f.write("\n")
                                    f.write(f"user: {str(result)}")
                                    f.write("\n")

                            else:
                                f.write(f"user: {str(enquiry)}")
                                f.write("\n")
                                f.write(f"assistant: {response}")
                                f.write("\n")
                    else:
                        f.write(f"user: {str(enquiry)}")
                        f.write("\n")
                        f.write(f"assistant: {response}")
                        f.write("\n")